from collections import UserDict
import json
from datetime import date, datetime

from psycopg2.extras import Json

# from django.core import exceptions
from django.utils.translation import ugettext_lazy as _
from django.contrib.postgres.fields import JSONField

from .models import Schema, compose_schemas


class JSONAttributes(UserDict):
    def __init__(self, *args, **kwargs):
        self._schemas = None
        self._instance = None
        self._setup = False
        super().__init__(*args, **kwargs)

    def setup_from_dict(self, dict):
        self.setup_schema()
        if dict is None or len(dict) == 0:
            self._setup = False
        else:
            for k, v in dict.items():
                self._check_key(k)
                self._attrs[k].validate(v)
                self[k] = v

    def setup_schema(self, schemas=None):
        if self._setup and schemas is None:
            return

        # Determine schemas for model instance containing this field.
        if schemas is not None:
            self._schemas = schemas
        else:
            self._schemas = Schema.objects.from_instance(self._instance)
        self._setup = True

        # Extract schema attributes, names of required attributes and
        # names of attributes with defaults, composing schemas for
        # instance.
        attrs = compose_schemas(*self._schemas)
        self._attrs, self._required_attrs, self._default_attrs = attrs

        # Fill in defaulted attributes.
        if len(self._required_attrs) > 0:
            for key in self._required_attrs:
                self[key] = self._attrs[key].default

    def _check_key(self, key):
        self.setup_schema()
        if key not in self._attrs:
            raise KeyError(key)

    def __getitem__(self, key):
        self._check_key(key)
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        self._check_key(key)
        self._attrs[key].validate(value)
        return super().__setitem__(key, value)

    def __delitem__(self, key):
        self._check_key(key)
        if key in self._required_attrs:
            raise KeyError(key)
        return super().__delitem__(key)

    @property
    def schemas(self):
        self.setup_schema()
        return self._schemas

    @property
    def attributes(self):
        self.setup_schema()
        return self._attrs


# This is needed to provide JSON serialisation for date objects
# whenever they're saved to JSON attribute fields.  This function is
# passed as the custom "dumps" method for psycopg2's Json class to
# use.

def json_serialiser(obj):
    return json.dumps(
        obj,
        default=lambda obj: (
            obj.isoformat()
            if isinstance(obj, datetime) or isinstance(obj, date)
            else None
        )
    )


class JSONAttributeField(JSONField):
    description = _('A managed JSON attribute set')

    def __init__(self, *args, **kwargs):
        kwargs['default'] = JSONAttributes
        super().__init__(*args, **kwargs)

    # def to_python(self, value):
    #     return JSONAttributes(value)

    def from_db_value(self, value, expression, connection, context):
        return value

    def get_prep_value(self, value):
        return (Json(dict(value), dumps=json_serialiser)
                if value is not None else value)

    def get_prep_lookup(self, lookup_type, value):
        if lookup_type in ('has_key', 'has_keys', 'has_any_keys'):
            return value
        return (Json(dict(value), dumps=json_serialiser)
                if isinstance(value, dict)
                else super().get_prep_lookup(lookup_type, value))

    # def validate(self, value, model_instance):
    #     super(JSONField, self).validate(value, model_instance)
    #     try:
    #         json.dumps(dict(value))
    #     except TypeError:
    #         raise exceptions.ValidationError(
    #             self.error_messages['invalid'],
    #             code='invalid',
    #             params={'value': value},
    #         )



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-06-11 06:32
from __future__ import unicode_literals

import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256)),
                ('long_name', models.CharField(blank=True, max_length=512)),
                ('index', models.IntegerField()),
                ('choices', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=256), null=True, size=None)),
                ('default', models.CharField(blank=True, max_length=256)),
                ('required', models.BooleanField(default=False)),
                ('omit', models.BooleanField(default=False)),
            ],
            options={
                'ordering': ('schema', 'index'),
            },
        ),
        migrations.CreateModel(
            name='AttributeType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256)),
                ('label', models.CharField(max_length=512)),
                ('form_field', models.CharField(max_length=256)),
                ('widget', models.CharField(blank=True, max_length=256, null=True)),
                ('validator_re', models.CharField(blank=True, max_length=512, null=True)),
                ('validator_type', models.CharField(blank=True, max_length=256, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Schema',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('selectors', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(blank=True, max_length=256), blank=True, default=list, size=None)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.AddField(
            model_name='attribute',
            name='attr_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='jsonattrs.AttributeType'),
        ),
        migrations.AddField(
            model_name='attribute',
            name='schema',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attributes', to='jsonattrs.Schema'),
        ),
        migrations.AlterUniqueTogether(
            name='schema',
            unique_together=set([('content_type', 'selectors')]),
        ),
        migrations.AlterUniqueTogether(
            name='attribute',
            unique_together=set([('schema', 'name'), ('schema', 'index')]),
        ),
    ]



from jsonattrs.models import Schema


class JsonAttrsMixin:
    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)

        obj = self.object
        field = self.attributes_field
        obj_attrs = getattr(obj, field)

        schemas = Schema.objects.from_instance(obj)
        attrs = [a for s in schemas for a in s.attributes.all()]
        context[field] = [(a.long_name, obj_attrs.get(a.name, '—'))
                          for a in attrs if not a.omit]
        return context



from collections import OrderedDict
import re

from django.db import models
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ValidationError
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import ArrayField


class SchemaManager(models.Manager):
    cache = dict()
    content_type_to_selectors = dict()

    @classmethod
    def invalidate_cache(cls):
        cls.cache = dict()

    def lookup(self, instance=None, content_type=None, selectors=None):
        if instance is not None and content_type is None:
            content_type = ContentType.objects.get_for_model(instance)

        if selectors is None and instance is not None:
            # Lazily pre-process per-content type selector definitions.
            if len(self.content_type_to_selectors) == 0:
                for k, v in settings.JSONATTRS_SCHEMA_SELECTORS.items():
                    a, m = k.split('.')
                    self.content_type_to_selectors[
                        ContentType.objects.get(app_label=a, model=m)
                    ] = v

            # Build full list of selectors from instance.
            selectors = []
            for s in self.content_type_to_selectors[content_type]:
                selector = instance
                for step in s.split('.'):
                    selector = getattr(selector, step, None)
                selectors.append(str(selector))
        selectors = tuple(selectors)
        if any(s is None for s in selectors):
            return None

        # Look for schema list in cache, keyed by content type and
        # selector list.
        key = (content_type, selectors)
        if key in self.cache:
            return self.cache[key]

        # Not in cache: build schema list using increasing selector
        # sequences.
        base_schemas = self.filter(content_type=content_type)
        schemas = []
        for i in range(len(selectors) + 1):
            schemas += list(base_schemas.filter(selectors=selectors[:i]))
        self.cache[key] = schemas
        return schemas

    def from_instance(self, instance):
        return self.lookup(instance=instance)


class Schema(models.Model):
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    selectors = ArrayField(models.CharField(max_length=256, blank=True),
                           default=list, blank=True)

    class Meta:
        unique_together = ('content_type', 'selectors')

    def __str__(self):
        return '<Schema #{}: {}+{}>'.format(
            self.id,
            self.content_type if self.content_type_id is not None else 'None',
            self.selectors
        )

    def __repr__(self):
        return str(self)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        SchemaManager.invalidate_cache()

    objects = SchemaManager()


def compose_schemas(*schemas):
    # Extract schema attributes, names of required attributes and
    # names of attributes with defaults, composing schemas.
    sattrs = [s.attributes.all() for s in schemas]
    attrs = OrderedDict()
    for sas in sattrs:
        for sa in sas:
            if sa.omit:
                if sa.name in attrs:
                    del attrs[sa.name]
            else:
                attrs[sa.name] = sa
    required_attrs = {n for n, a in attrs.items() if a.required}
    default_attrs = {n for n, a in attrs.items() if a.default is not None}

    return attrs, required_attrs, default_attrs


def create_attribute_type(name, label, form_field,
                          widget=None, validator_re=None, validator_type=None):
    AttributeType.objects.create(
        name=name, label=label, form_field=form_field, widget=widget,
        validator_re=validator_re, validator_type=validator_type
    )


class AttributeType(models.Model):
    name = models.CharField(max_length=256)
    label = models.CharField(max_length=512)
    form_field = models.CharField(max_length=256)
    widget = models.CharField(max_length=256, null=True, blank=True)
    validator_re = models.CharField(max_length=512, null=True, blank=True)
    validator_type = models.CharField(max_length=256, null=True, blank=True)


def create_attribute_types():
    create_attribute_type('boolean', 'Boolean', 'BooleanField',
                          validator_type='bool',
                          validator_re=r'true|false|True|False')

    create_attribute_type('text', 'Text', 'CharField',
                          validator_type='str')
    create_attribute_type('text_multiline', 'Multiline text', 'CharField',
                          validator_type='str',
                          widget='Textarea')

    create_attribute_type('date', 'Date', 'DateField')
    create_attribute_type('dateTime', 'Date and time', 'DateTimeField')
    create_attribute_type('time', 'Time', 'TimeField')

    create_attribute_type('integer', 'Integer', 'IntegerField',
                          validator_re=r'[-+]?\d+')
    create_attribute_type('decimal', 'Decimal number', 'DecimalField',
                          validator_re=r'[-+]?\d+(\.\d+)?')

    create_attribute_type('email', 'Email address', 'EmailField')
    create_attribute_type('url', 'URL', 'URLField')

    create_attribute_type('select_one', 'Select one:', 'ChoiceField')
    create_attribute_type('select_multiple', 'Select multiple:',
                          'MultipleChoiceField')
    create_attribute_type('foreign_key', 'Select one:', 'ModelChoiceField')


# Sigh.  Oh Python.  Really?  Why?  Is there really no other way to
# look up a class by name than to manually calculate the transitive
# closure of the subclass relationship starting from "object"?

def desc_classes(c):
    try:
        scs = c.__subclasses__()
        sscs = []
        for sc in scs:
            sscs += desc_classes(sc)
        scs += sscs
        return scs
    except:
        return []


class_cache = None


def find_class(name):
    global class_cache

    if class_cache is None:
        class_cache = desc_classes(object)
    return [c for c in class_cache if c.__name__ == name][0]


class Attribute(models.Model):
    schema = models.ForeignKey(
        Schema, related_name='attributes', on_delete=models.CASCADE
    )
    name = models.CharField(max_length=256)
    long_name = models.CharField(max_length=512, blank=True)
    attr_type = models.ForeignKey(AttributeType, on_delete=models.CASCADE)
    index = models.IntegerField()
    choices = ArrayField(models.CharField(max_length=256), null=True)
    default = models.CharField(max_length=256, blank=True)
    required = models.BooleanField(default=False)
    omit = models.BooleanField(default=False)

    class Meta:
        ordering = ('schema', 'index')
        unique_together = (('schema', 'index'), ('schema', 'name'))

    def validate(self, value):
        if self.choices is not None and self.choices != []:
            if value not in self.choices:
                raise ValidationError(
                    _('Invalid choice for %(field)s: "%(value)s"'),
                    params={'field': self.name, 'value': value}
                )
        atype = self.attr_type
        if isinstance(value, str):
            if (atype.validator_re is not None and
               re.match(atype.validator_re, value) is None):
                raise ValidationError(
                    _('Validation failed for %(field)s: "%(value)s"'),
                    params={'field': self.name, 'value': value}
                )
        elif atype.validator_type is not None:
            if not isinstance(value, find_class(atype.validator_type)):
                raise ValidationError(
                    _('Validation failed for %(field)s: "%(value)s"'),
                    params={'field': self.name, 'value': value}
                )

    def __str__(self):
        return '<{} attribute {}: required={} default={}>'.format(
            self.attr_type.form_field,
            self.name,
            self.required, self.default
        )

    def __repr__(self):
        return str(self)



from django import forms
from django.contrib.contenttypes.models import ContentType

from .models import Schema, compose_schemas


FORM_FIELDS = {
    'BooleanField': forms.BooleanField,
    'CharField': forms.CharField,
    'IntegerField': forms.IntegerField,
    'DecimalField': forms.DecimalField,
    'DateField': forms.DateField,
    'DateTimeField': forms.DateTimeField,
    'TimeField': forms.TimeField,
    'EmailField': forms.EmailField,
    'URLField': forms.URLField,
    'ChoiceField': forms.ChoiceField,
    'MultipleChoiceField': forms.MultipleChoiceField
}


def form_field_from_name(name):
    if name in FORM_FIELDS:
        return FORM_FIELDS[name]
    else:
        raise ValueError("Invalid form field name "
                         "for attribute: '{}'".format(name))


class AttributeModelForm(forms.ModelForm):
    attributes_field = None

    def __init__(self, *args, **kwargs):
        schema_selectors = kwargs.pop('schema_selectors', None)
        super().__init__(*args, **kwargs)
        if self.attributes_field is not None:
            self.add_attribute_fields(schema_selectors)

    def add_attribute_fields(self, schema_selectors):
        attrs = None
        attrvals = getattr(self.instance, self.attributes_field)
        schemas = None
        if self.instance.pk:
            schemas = Schema.objects.from_instance(self.instance)
        elif schema_selectors is not None:
            selectors = []
            for ss in schema_selectors:
                selectors.append(ss['selector'])
                if ss['name'] is not None:
                    setattr(self.instance, ss['name'], ss['value'])
            content_type = ContentType.objects.get_for_model(self.Meta.model)
            schemas = Schema.objects.lookup(
                content_type=content_type, selectors=selectors
            )
            attrvals.setup_schema(schemas)
        attrs, _, _ = compose_schemas(*schemas)
        for name, attr in attrs.items():
            fieldname = self.attributes_field + '::' + name
            atype = attr.attr_type
            args = {'label': attr.long_name, 'required': False}
            field = form_field_from_name(atype.form_field)
            if atype.form_field == 'CharField':
                args['max_length'] = 32
            if (atype.form_field == 'ChoiceField' or
               atype.form_field == 'MultipleChoiceField'):
                args['choices'] = list(map(lambda c: (c, c), attr.choices))
            if atype.form_field == 'BooleanField':
                args['required'] = False
                self.set_default(args, attr, boolean=True)
            elif attr.required:
                args['required'] = True
                self.set_default(args, attr)
            else:
                self.set_default(args, attr)
            self.set_initial(args, name, attr, attrvals)
            self.fields[fieldname] = field(**args)

    def set_default(self, args, attr, boolean=False):
        if len(attr.default) > 0:
            if boolean:
                args['initial'] = (attr.default != 'False')
            else:
                args['initial'] = attr.default

    def set_initial(self, args, name, attr, attrvals):
        if name in attrvals:
            if attr.attr_type.form_field == 'BooleanField':
                if isinstance(attrvals[name], bool):
                    args['initial'] = attrvals[name]
                else:
                    args['initial'] = attrvals[name] != 'False'
            else:
                args['initial'] = attrvals[name]

    def process_attributes_fields(self):
        chk = self.attributes_field + '::'
        chklen = len(chk)
        attrvals = getattr(self.instance, self.attributes_field)
        for k, v in self.cleaned_data.items():
            if k.startswith(chk):
                k = k[chklen:]
                attrvals[k] = v
        setattr(self.instance, self.attributes_field, attrvals)

    def save(self, *args, **kwargs):
        if self.attributes_field is not None:
            self.process_attributes_fields()
        return super().save(*args, **kwargs)



from django.core.management.base import BaseCommand

from jsonattrs.models import AttributeType, create_attribute_types


def run():
    AttributeType.objects.all().delete()
    create_attribute_types()


class Command(BaseCommand):
    help = "Load default attribute types."

    def handle(self, *args, **options):
        run()



from django.db.models.signals import post_init

from .fields import JSONAttributes, JSONAttributeField


def fixup_instance(sender, **kwargs):
    instance = kwargs['instance']
    for f in instance._meta.fields:
        if isinstance(f, JSONAttributeField):
            fi = getattr(instance, f.name)
            if not isinstance(fi, JSONAttributes):
                setattr(instance, f.name, JSONAttributes())
            getattr(instance, f.name)._instance = instance
            getattr(instance, f.name).setup_from_dict(fi)


def fix_model_for_attributes(cls):
    post_init.connect(fixup_instance, sender=cls)
    return cls
